from coinor.pulp import *
import coinor.dippy as dippy

##import sys
import platform

doRelaxed = True
doCustomRelaxed = True
writeFiles = not doRelaxed

length = {
"9cm": 9,
"7cm": 7,
"5cm": 5
##"13cm": 13
}

ITEMS = length.keys()

demand = {
"9cm": 1,
"7cm": 1,
"5cm": 4
##"9cm": 1,
##"7cm": 1,
##"5cm": 0
##"13cm": 1
##"9cm": 3,
##"7cm": 4,
##"5cm": 6
##"9cm": 150,
##"7cm": 200,
##"5cm": 300
}

total_patterns = sum(demand[i] for i in ITEMS)

total_length = {}
for p in range(total_patterns):
    total_length[p] = 20
PATTERNS = total_length.keys()

CUTS = [(p, i) for p in PATTERNS
               for i in ITEMS]

prob = dippy.DipProblem("Sponge_Rolls", LpMinimize)

# create variables

useVars = LpVariable.dicts("Use", PATTERNS, 0, 1, LpBinary)

cutVars = LpVariable.dicts("Cut", CUTS, 0, 10, LpInteger)

# objective
prob += sum(useVars[p] for p in PATTERNS), "min"

# Meet demand
for i in ITEMS:
    prob += sum(cutVars[(p, i)] for p in PATTERNS) \
            >= demand[i]

# Ordering patterns
for i, p in enumerate(PATTERNS):
    if p != PATTERNS[-1]:
         prob += useVars[p] >= useVars[PATTERNS[i+1]]

# Cut patterns
for p in PATTERNS:
    if doRelaxed:
        prob.relaxation[p] += sum(length[i] *
                                  cutVars[(p, i)] for i in ITEMS) \
                                  <= total_length[p] * useVars[p]
    else:
        prob += sum(length[i] * cutVars[(p, i)] for i in ITEMS) \
             <= total_length[p] * useVars[p]

V = []
A = []

def printNodeInfo(prob, info):
    nodeInd   = info["nodeIndex"]
    parentInd = info["parentIndex"]

    if parentInd != -1:
        V.append(nodeInd)
        A.append((parentInd, nodeInd))

prob.post_process = printNodeInfo

def relaxed_solver(prob, patt, redCosts, convexDual):
##    print patt, "in", PATTERNS
##    print "redCosts =", redCosts
##    print "convexDual =", convexDual
    # get items with negative reduced cost
    item_idx = [i for i in ITEMS \
                if redCosts[cutVars[(patt, i)]] < 0]
    vars = [cutVars[(patt, i)] for i in item_idx]
    obj = [-redCosts[cutVars[(patt, i)]] for i in item_idx]
    weights = [length[i] for i in item_idx]

##    print "Using knapsack heuristic"
##    print "item_idx =", item_idx
##    print "obj =", obj
##    print "weights =", weights
##    print "capacity =", total_length[patt]
    z, solution = kp(obj, weights, total_length[patt])
##    print "Number of items = ", len(item_idx)
##    for i in range(len(item_idx)):
##        print "Item ", item_idx[i], " has profit ", obj[i], " and weight ", weights[i]
##    print "Knapsack has capacity ", total_length[patt]
##    print "Value = ", z
##    for i in range(len(item_idx)):
##        print "Included[", item_idx[i], "] = ", solution[i]
    
    total_weight = sum(w * solution[i] \
                       for i, w in enumerate(weights))
    assert total_weight <= total_length[patt]

    # add in reduced cost of useVars
    totalCut = sum(solution)
##    print "z, redCosts[useVars[", patt, "]], convexDual", z, redCosts[useVars[patt]], convexDual
    if totalCut > 0:
        rc = -z + redCosts[useVars[patt]] - convexDual
    else:
        rc = -convexDual
        
##    print "rc =", rc
##    sys.stdout.flush()
    if rc < 0: # Using this pattern
        var_values = dict([(v, solution[i]) \
                           for i, v in enumerate(vars) \
                           if solution[i] > 0])
        if totalCut > 0:
            var_values[useVars[patt]] = 1
            cost = 1
        else:
            cost = 0

        var_tuple = (cost, rc, var_values)
        return [var_tuple]
    
    return []

if doRelaxed and doCustomRelaxed:
    prob.relaxed_solver = relaxed_solver

def kp(obj, weights, capacity):
    assert len(obj) == len(weights)
    n = len(obj)

    if n == 0:
        return 0, []

    if capacity == 0:
        return 0, [0 for i in range(n)]
    
    n = len(obj)

    # Don't include item
    zbest, solbest = kp(obj, weights, capacity - 1)
    # Check all items for inclusion
    for i in range(n):
        if weights[i] <= capacity:
            zyes, solyes = kp(obj, weights, \
                              capacity - weights[i])
            zyes += obj[i]
            solyes[i] += 1
            if zyes > zbest:
                zbest = zyes
                solbest = solyes

    return zbest, solbest

##def knapsack_integer(obj, weights, capacity):
##    assert len(obj) == len(weights)
##    n = len(obj)
##
##    if n == 0:
##        return 0, []
##
##    c = [[0]*(capacity+1) for i in range(n)]
##    added = [[False]*(capacity+1) for i in range(n)]
##
##    # c[i,j] is the optimal solution for i products, j capacity
##
##    for i in range(n):
##        for j in range(capacity+1):
##            # either, weight of item doesn't fit in remaining cap
##            if weights[i] > j:
##                c[i][j] = c[i-1][j]
##            elif i == 0:
##                # special case for item 0, 
##                # always add a new item
##                c[i][j] = obj[i] + c[i][j-weights[i]]
##                added[i][j] = True
##            else:
##                # choice of adding another of this item
##                # compared to not having any of this item
##                c_add = obj[i] + c[i][j-weights[i]]
##                if c_add > c[i-1][j]:
##                    c[i][j] = c_add
##                    added[i][j] = True
##                else:
##                    c[i][j] = c[i-1][j]
##
##    # backtrack to find solution
##    i = n-1
##    j = capacity
##    solution = [0] * n
##    while i >= 0 and j >= 0:
##        if added[i][j]:
##            solution[i] += 1
##            j -= weights[i]
##        else:
##            i -= 1
##
##    return c[n-1][capacity], solution
##
### test knapsack solver
##o = [10, 1, 3]
##w = [1, 9, 9]
##z, sol = knapsack_integer(o, w, 10)
##assert z == 100
##assert sol == [10, 0, 0]
##
##o = [6, 4, 3]
##w = [4, 3, 3]
##z, sol = knapsack_integer(o, w, 10)
##assert z == 14
##assert sol == [1, 2, 0]

##prob.writeLP("patterns.lp")
##for p in PATTERNS:
##    prob.writeRelaxed(p, "pattCons_%s.lp" % p)

if writeFiles:
    prob.writeMPS("cutting_stock1.mps")
    prob.writeLP("cutting_stock1.lp")

if doRelaxed and (len(prob.relaxation.keys()) > 0):
    dippy.Solve(prob, {
        'doPriceCut':1,
##        'LogLevel': 0,
##        'LogDebugLevel': 5,
##        'LogDumpModel': 5,
        'CutCGL': 0,
    })
else:
    dippy.Solve(prob, {
##        'LogDebugLevel': 5,
##        'LogDumpModel': 5,
    })

for i, var in useVars.items():
    if var.varValue:
        print "Use", i, var.varValue

for pat in PATTERNS:
    for i in ITEMS:
        if cutVars[(pat, i)].varValue:
            print "Pat", pat, "item", i, \
                  cutVars[(pat, i)].varValue

##for (pat, w), var in cutVars.items():
##    if var.varValue:
##        print "Pat", pat, "item", w, var.varValue

from utils import visualise

visualise(V, A)


